import { IArrayFormat } from "@daybrush/utils";
import { SceneItem } from ".";
export interface EasingFunction extends Function {
    easingName?: string;
}
export declare type FillModeType = "forwards" | "backwards" | "both";
export declare type IterationCountType = number | "infinite";
export declare type EasingType = 0 | EasingFunction;
export declare type DirectionType = "normal" | "reverse" | "alternate" | "alternate-reverse";
export declare type PlayStateType = "paused" | "running";
export interface AnimatorOptions {
    id: number | string;
    easing: string | EasingType;
    iterationCount: IterationCountType;
    delay: number;
    fillMode: FillModeType;
    direction: DirectionType;
    playSpeed: number;
    duration: number;
}
export interface AnimatorState extends AnimatorOptions {
    easingName: string;
    iterationTime: number;
    currentTime: number;
    tickTime: number;
    iteration: number;
    prevTime: number;
    playState: PlayStateType;
}
export interface SceneState extends AnimatorState {
    selector: string | boolean | ((id: number | string) => string);
    playCSS: boolean;
    exportEvent?: boolean;
}
export interface SceneOptions extends AnimatorOptions {
    selector: string | boolean | ((id: number | string) => string);
}
export interface SceneItemState extends AnimatorState {
    playCSS: boolean;
    cssText: string;
    selector: string;
    exportEvent?: boolean;
}
export interface SceneItemOptions extends AnimatorOptions {
    selector: string | boolean | ((id: number | string) => string);
    elements: IArrayFormat<AnimateElement> | AnimateElement;
    element: IArrayFormat<AnimateElement> | AnimateElement;
    target: any;
}
export interface PropertyObjectState {
    prefix: string;
    suffix: string;
    model: string;
    type: string;
    separator: string;
}
export declare type Role = true | RoleObject;
export declare type ElementsType = HTMLElement | HTMLElement[] | NodeListOf<HTMLElement>;
export declare type CallbackType<T = any> = (...args: any[]) => T;
export interface EventParameter {
    [name: string]: CallbackType | CallbackType[];
}
export interface RoleObject {
    [role: string]: Role;
}
export declare type NameType = string | number;
export declare type OptionType = ["duration", "fillMode", "direction", "iterationCount", "delay", "easing", "playSpeed"];
export declare type EventType = ["paused", "ended", "timeupdate", "animate", "play", "iteration"];
export declare type AnimateElement = HTMLElement | SVGElement;
export interface PlayCondition {
    className?: string;
    selector?: string | ((item: SceneItem, selector: string) => string);
}
